# Classes
from vision_models_playground.models.classifiers.conv_vision_transformer import ConvVisionTransformer
from vision_models_playground.models.classifiers.resnet import ResNet
from vision_models_playground.models.classifiers.vision_transformer import VisionTransformer
from vision_models_playground.models.classifiers.ff_classifier import FeedForwardClassifier
from vision_models_playground.models.classifiers.perceiver import Perceiver
from vision_models_playground.models.classifiers.conv_classifier import ConvolutionalClassifier
from vision_models_playground.models.classifiers.vision_perceiver import VisionPerceiver
# Functions
from vision_models_playground.models.classifiers.conv_vision_transformer import build_cvt_13
from vision_models_playground.models.classifiers.conv_vision_transformer import build_cvt_21
from vision_models_playground.models.classifiers.conv_vision_transformer import build_cvt_w24
from vision_models_playground.models.classifiers.resnet import build_resnet_18
from vision_models_playground.models.classifiers.resnet import build_resnet_34
from vision_models_playground.models.classifiers.resnet import build_resnet_50
from vision_models_playground.models.classifiers.resnet import build_resnet_101
from vision_models_playground.models.classifiers.resnet import build_resnet_152



from typing import Union, List, Optional, Callable

from torch import nn

from einops import rearrange

from vision_models_playground.components.position_embedding import FourierEmbedding
from vision_models_playground.models.classifiers.perceiver import Perceiver
from vision_models_playground.utility.datasets import get_cifar10_dataset
from vision_models_playground.utility.train_models import train_model_classifier


class VisionPerceiver(nn.Module):
    def __init__(
            self,
            *,
            patch_size: int,
            projection_dim: int,

            num_classes: int = 1000,

            apply_rotary_emb: bool = True,

            apply_fourier_encoding: bool = True,
            max_freq: int = 10,
            num_freq_bands: int = 6,
            constant_mapping: bool = False,
            max_position: Union[List[int], int] = 1600,

            num_layers: int = 2,

            num_latents: int = 128,
            latent_dim: int = 512,

            cross_num_heads: int = 8,
            cross_head_dim: int = 64,

            self_attend_heads: int = 8,
            self_attend_dim: int = 64,

            transformer_depth: int = 2,

            attention_dropout: float = 0.0,
            ff_hidden_dim: int = 512,
            ff_dropout: float = 0.0,
            activation: Optional[Callable] = None,
    ):
        super().__init__()
        patch_height, patch_width = patch_size, patch_size

        self.patch_height, self.patch_width = patch_height, patch_width

        # Create fourier encoder
        self.pos_encoder = FourierEmbedding(max_freq, num_freq_bands, constant_mapping, max_position) \
            if apply_fourier_encoding else nn.Identity()

        self.to_projection = nn.LazyLinear(projection_dim)

        # build perceiver
        self.perceiver = Perceiver(
            input_dim=projection_dim,
            input_axis=1,
            final_classifier_head=True,
            num_classes=num_classes,
            apply_rotary_emb=apply_rotary_emb,
            apply_fourier_encoding=False,
            num_layers=num_layers,
            num_latents=num_latents,
            latent_dim=latent_dim,
            cross_num_heads=cross_num_heads,
            cross_head_dim=cross_head_dim,
            self_attend_heads=self_attend_heads,
            self_attend_dim=self_attend_dim,
            transformer_depth=transformer_depth,
            attention_dropout=attention_dropout,
            ff_hidden_dim=ff_hidden_dim,
            ff_dropout=ff_dropout,
            activation=activation,
        )

    def forward(self, x):
        x = rearrange(x, 'b c h w -> b h w c')
        x = self.pos_encoder(x)
        x = rearrange(x, 'b (h p1) (w p2) c -> b (h w) (p1 p2 c)', p1=self.patch_height, p2=self.patch_width)
        x = self.to_projection(x)
        x = self.perceiver(x)

        return x


def main():
    model = VisionPerceiver(
        patch_size=4,
        projection_dim=1024,
        num_classes=10,
        apply_rotary_emb=True,
        max_position=1600,
        num_layers=2,
        num_latents=16,
        latent_dim=32,
        cross_num_heads=8,
        cross_head_dim=64,
        self_attend_heads=8,
        self_attend_dim=64,
        transformer_depth=2,
        attention_dropout=0.0,
        ff_hidden_dim=512,
        ff_dropout=0.0,
        activation=None,
    ).cuda()

    train_dataset, test_dataset = get_cifar10_dataset()
    train_model_classifier(model, train_dataset, test_dataset, batch_size=128)


if __name__ == "__main__":
    main()



import torch
from torch import nn


class QuickGELU(nn.Module):
    """
    This is a quick GELU implementation that is faster than the official
    implementation.

    This implementation is based on the following paper:
    https://arxiv.org/abs/1606.08415
    """

    def forward(self, x: torch.Tensor):
        return x * torch.sigmoid(1.702 * x)


def main():
    import matplotlib.pyplot as plt

    x = torch.linspace(-5, 5, 100)

    # Compute the GELU values
    y_gelu = torch.nn.GELU()(x)

    # Compute the QuickGELU values
    y_quick_gelu = QuickGELU()(x)

    # Plot the results
    plt.plot(x.numpy(), y_gelu.numpy(), label="GELU")
    plt.plot(x.numpy(), y_quick_gelu.numpy(), label="QuickGELU")
    plt.legend()
    plt.show()


if __name__ == "__main__":
    main()



from typing import Optional, Callable, Literal

import torch
from torch import nn

from vision_models_playground.components.attention.transformer_decoder import TransformerDecoder
from vision_models_playground.components.attention.transformer_encoder import TransformerEncoder


class Transformer(nn.Module):
    def __init__(
            self,
            target_dim: int,
            context_dim: int,
            depth: int,
            heads: int,
            head_dim: int,
            mlp_dim: int,
            mlp_dropout: float = 0.0,
            attention_dropout: float = 0.0,
            apply_rotary_emb: bool = False,
            activation: Optional[Callable] = None,
            drop_path: float = 0.0,
            norm_type: Literal['pre_norm', 'post_norm'] = 'pre_norm',
    ):
        """
        Args:
            target_dim:
                Dimension of the input tensor.

            context_dim:
                Dimension of the input tensor.

            depth:
                Number of layers in the encoder.

            heads:
                Number of attention heads.

            head_dim:
                Dimension of each attention head.

            mlp_dim:
                Dimension of the MLP.

            mlp_dropout:
                Dropout probability for the MLP.

            attention_dropout:
                Dropout probability for the attention.

            apply_rotary_emb:
                Whether to apply rotary embedding to the queries.

            activation:
                Activation function for the MLP.

            drop_path:
                Dropout probability for the drop path.

            norm_type:
                Type of normalization to use.
        """

        super().__init__()

        self.encoder = TransformerEncoder(
            dim=context_dim,
            depth=depth,
            heads=heads,
            head_dim=head_dim,
            mlp_dim=mlp_dim,
            mlp_dropout=mlp_dropout,
            attention_dropout=attention_dropout,
            apply_rotary_emb=apply_rotary_emb,
            activation=activation,
            drop_path=drop_path,
            norm_type=norm_type,
        )

        self.decoder = TransformerDecoder(
            target_dim=target_dim,
            context_dim=context_dim,
            depth=depth,
            heads=heads,
            head_dim=head_dim,
            mlp_dim=mlp_dim,
            mlp_dropout=mlp_dropout,
            attention_dropout=attention_dropout,
            apply_rotary_emb=apply_rotary_emb,
            activation=activation,
            drop_path=drop_path,
            norm_type=norm_type,
        )

    def forward(
            self,
            target: torch.Tensor,
            context: torch.Tensor,
            target_mask: Optional[torch.Tensor] = None,
            context_mask: Optional[torch.Tensor] = None,
            return_context_too: bool = False,
            causal: bool = False,
    ) -> torch.Tensor:
        """
        Args:
            target:
                Target tensor. [batch_size, seq_len_1, dim_1]

            context:
                Context tensor. [batch_size, seq_len_2, dim_2]

            target_mask:
                Target mask. [batch_size, seq_len_1]

            context_mask:
                Context mask. [batch_size, seq_len_2]

            return_context_too:
                Whether to return the context.

            causal:
                Whether to use causal attention to the target.

        Returns:
            Output tensor. [batch_size, seq_len_1, dim_1] if return_context_too is False,
            otherwise Tuple[Tensor[batch_size, seq_len_1, seq_len_1, dim_1], Tensor[batch_size, seq_len_2, dim_2]]
        """

        # Compute the masks
        encoder_self_attention_mask = self.encoder.compute_2d_mask(context_mask)

        decoder_self_attention_mask = self.decoder.compute_2d_mask(target_mask)
        decoder_self_attention_mask = self.decoder.compute_causal_mask(decoder_self_attention_mask, causal)

        decoder_cross_attention_mask = self.decoder.compute_2d_mask(context_mask, target.shape[1])

        # Forward pass
        for encoder_layer, decoder_layer in zip(self.encoder.layers, self.decoder.layers):
            context = encoder_layer(context, encoder_self_attention_mask)
            target = decoder_layer(target, context, decoder_self_attention_mask, decoder_cross_attention_mask)

        return target if not return_context_too else (target, context)


def main():
    model = Transformer(
        target_dim=64,
        context_dim=128,
        depth=6,
        heads=8,
        head_dim=64,
        mlp_dim=2048,
        mlp_dropout=0.1,
        attention_dropout=0.1,
        apply_rotary_emb=True,
        activation=nn.ReLU(),
        drop_path=0.1,
        norm_type='pre_norm',
    )

    target = torch.randn(2, 6, 64)
    context = torch.randn(2, 10, 128)

    target_mask = torch.ones(2, 6).bool()
    target_mask[:, 3:] = 0
    context_mask = torch.ones(2, 10).bool()
    context_mask[:, 8:] = 0

    output = model(target, context, target_mask, context_mask)
    print(output.shape)


if __name__ == '__main__':
    main()



from typing import Optional

import torch
from einops import rearrange, repeat
from rotary_embedding_torch import RotaryEmbedding
from torch import nn, einsum

from vision_models_playground.utility.functions import default, exists


class Attention(nn.Module):
    def __init__(
            self,
            query_dim: int,
            context_dim: Optional[int] = None,
            heads: int = 8,
            dim_head: int = 64,
            dropout: float = 0.0,
            apply_rotary_emb: bool = False
    ):
        super(Attention, self).__init__()

        inner_dim = dim_head * heads
        context_dim = default(context_dim, query_dim)

        self.scale = dim_head ** -0.5
        self.heads = heads

        self.to_q = nn.Linear(query_dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(context_dim, inner_dim * 2, bias=False)

        self.dropout = nn.Dropout(dropout)
        self.to_out = nn.Linear(inner_dim, query_dim)

        self.rotary_emb = None
        if apply_rotary_emb:
            assert dim_head % 2 == 0, "Rotary embedding cannot be applied on odd dimensions"
            self.rotary_emb = RotaryEmbedding(dim=dim_head // 2)

    def forward(
            self,
            queries: torch.Tensor,
            context: Optional[torch.Tensor] = None,
            mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        # Get useful info
        h = self.heads
        context = default(context, queries)

        # Get query, key and value
        q = self.to_q(queries)
        k, v = self.to_kv(context).chunk(2, dim=-1)

        # Reshape query, key and value according to heads
        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=h), [q, k, v])

        # Apply rotary embedding if necessary
        if exists(self.rotary_emb):
            q, k = map(lambda t: self.rotary_emb.rotate_queries_or_keys(t), [q, k])

        # Apply attention
        sim = einsum("b h i d, b h j d -> b h i j", q, k) * self.scale

        # Mask attention if necessary
        if exists(mask):
            mask = repeat(mask, "b ... -> b h ...", h=h)

            max_neg_value = -torch.finfo(sim.dtype).max
            sim.masked_fill_(~mask, max_neg_value)

        # Compute attention sim matrix
        attn = sim.softmax(dim=-1)
        attn = self.dropout(attn)

        # Apply attention to values
        out = einsum("b h i j, b h j d -> b h i d", attn, v)
        out = rearrange(out, "b h n d -> b n (h d)")

        # Project to output channels
        return self.to_out(out)



import vision_models_playground.models.augmenters
import vision_models_playground.models.autoencoders
import vision_models_playground.models.classifiers
import vision_models_playground.models.generative
import vision_models_playground.models.segmentation


# Classes
from vision_models_playground.components.activations.relu_squared import ReluSquared
from vision_models_playground.components.activations.swish import Swish
from vision_models_playground.components.activations.geglu import GEGLU
from vision_models_playground.components.activations.mish import Mish
from vision_models_playground.components.activations.hard_sigmoid import HardSigmoid
from vision_models_playground.components.activations.quick_gelu import QuickGELU
from vision_models_playground.components.activations.quick_geglu import QuickGEGLU



from typing import Literal

import torch
from einops import rearrange
from torch import nn, einsum

from vision_models_playground.components.convolutions.conv_projection import ConvProjection


class ConvAttention(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            num_heads: int,
            dim_per_head: int = 64,
            qkv_bias: bool = False,
            attn_drop: float = 0.0,
            output_drop: float = 0.0,
            kernel_size: int = 3,
            stride_kv: int = 1,
            stride_q: int = 1,
            padding_kv: int = 1,
            padding_q: int = 1,
            method: Literal['conv', 'avg', 'linear'] = 'conv',
    ):
        super().__init__()

        # Save params
        self.stride_kv = stride_kv
        self.stride_q = stride_q
        self.num_heads = num_heads
        self.scale = dim_per_head ** -0.5
        self.inner_dim = dim_per_head * num_heads

        method_q = 'linear' if method == 'avg' else method
        method_kw = method

        # Get convolution projections components for query, key and value
        self.to_q = ConvProjection(
            in_channels, self.inner_dim, kernel_size, padding_q, stride_q, method_q, qkv_bias,
        )
        self.to_k = ConvProjection(
            in_channels, self.inner_dim, kernel_size, padding_kv, stride_kv, method_kw, qkv_bias
        )
        self.to_v = ConvProjection(
            in_channels, self.inner_dim, kernel_size, padding_kv, stride_kv, method_kw, qkv_bias
        )

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(self.inner_dim, out_channels)
        self.proj_drop = nn.Dropout(output_drop)

    def forward(self, x):
        # Get query, key and value
        q, k, v = self.to_q(x), self.to_k(x), self.to_v(x)

        # Reshape query, key and value according to heads
        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.num_heads), [q, k, v])

        # Calculate attention sim matrix
        attn_sim = einsum("b h i d, b h j d -> b h i j", q, k) * self.scale
        attn = attn_sim.softmax(dim=-1)
        attn = self.attn_drop(attn)

        # Apply attention to values
        out = einsum("b h i j, b h j d -> b h i d", attn, v)
        out = rearrange(out, "b h n d -> b n (h d)")

        # Project to output channels
        out = self.proj(out)
        out = self.proj_drop(out)

        return out


def main():
    block = ConvAttention(in_channels=3, out_channels=60, num_heads=8)
    x = torch.randn(1, 3, 64, 64)
    print(block(x).shape)


if __name__ == "__main__":
    main()


from typing import Union, List, Callable, Optional

import torch
from einops import rearrange, repeat
from einops.layers.torch import Reduce, Rearrange
from torch import nn

from vision_models_playground.components.activations.geglu import GEGLU
from vision_models_playground.components.attention import TransformerEncoder, FeedForward
from vision_models_playground.components.attention.attend import Attend
from vision_models_playground.components.position_embedding import FourierEmbedding
from vision_models_playground.utility.datasets import get_cifar10_dataset
from vision_models_playground.utility.train_models import train_model_classifier


class Perceiver(nn.Module):
    def __init__(
            self,
            input_dim: int,
            input_axis: int,

            final_classifier_head: bool = True,
            num_classes: int = 1000,

            apply_rotary_emb: bool = True,

            apply_fourier_encoding: bool = True,
            max_freq: int = 10,
            num_freq_bands: int = 6,
            constant_mapping: bool = False,
            max_position: Union[List[int], int] = 1600,

            num_layers: int = 2,

            num_latents: int = 128,
            latent_dim: int = 512,

            cross_num_heads: int = 8,
            cross_head_dim: int = 64,

            self_attend_heads: int = 8,
            self_attend_dim: int = 64,

            transformer_depth: int = 2,

            attention_dropout: float = 0.0,
            ff_hidden_dim: int = 512,
            ff_dropout: float = 0.0,
            activation: Optional[Callable] = None,
    ):
        """
        Initialize the Perceiver
        """

        super().__init__()

        # Default activation is GEGLU
        if activation is None:
            activation = GEGLU()

        # Create latent space
        self.latents = nn.Parameter(torch.randn(num_latents, latent_dim))
        self.latent_dim = latent_dim

        # Create fourier encoder
        self.pos_encoder = FourierEmbedding(max_freq, num_freq_bands, constant_mapping, max_position) \
            if apply_fourier_encoding else nn.Identity()

        # Compute the pos_encoder output dim
        self.pos_encoder_dim = input_dim + input_axis * ((num_freq_bands * 2) + 1) if apply_fourier_encoding else input_dim

        def create_cross_attend():
            return Attend(
                query_dim=latent_dim,
                context_dim=self.pos_encoder_dim,
                num_heads=cross_num_heads,
                head_dim=cross_head_dim,
                attention_dropout=attention_dropout,
                apply_rotary_emb=apply_rotary_emb,
                ff_hidden_dim=ff_hidden_dim,
                ff_dropout=ff_dropout,
                activation=activation
            )

        def create_latent_transformer():
            return TransformerEncoder(
                dim=latent_dim,
                depth=transformer_depth,
                heads=self_attend_heads,
                head_dim=self_attend_dim,
                mlp_dim=ff_hidden_dim,
                mlp_dropout=ff_dropout,
                attention_dropout=attention_dropout,
                apply_rotary_emb=apply_rotary_emb
            )

        # Build components
        self.layers = nn.ModuleList([])
        for i in range(num_layers):
            layer = nn.ModuleList([])

            layer.append(create_cross_attend())
            layer.append(create_latent_transformer())

            self.layers.append(layer)

        # Create output layer
        self.classifier = nn.Sequential(
            Reduce("b l d -> b d", "mean"),
            nn.LayerNorm(latent_dim),
            FeedForward(
                dim=latent_dim,
                hidden_dim=ff_hidden_dim,
                dropout=0.0,
                activation=activation,
                output_dim=num_classes
            )
        ) if final_classifier_head else nn.Identity()

    def forward(
            self,
            x: torch.Tensor,
            mask: Optional[torch.BoolTensor] = None,
            latents: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # Take input about the shape of the input
        batch_size = x.shape[0]

        # Initialize Latents
        latents = self.latents if latents is None else latents
        assert latents.shape[-1] == self.latent_dim

        # Repeat latents for each batch
        if len(latents.shape) == 2:
            latents = repeat(latents, "l d -> b l d", b=batch_size)

        # Apply position encoding
        x = self.pos_encoder(x)

        # Flatten the input
        x = rearrange(x, "b ... d -> b (...) d")

        for cross_attend, transformer in self.layers:
            latents = cross_attend(latents, x, mask)
            latents = transformer(latents)

        output = self.classifier(latents)
        return output


def main():
    model = Perceiver(
        input_dim=1,
        input_axis=2,
        final_classifier_head=True,
        num_classes=10,
        apply_rotary_emb=True,
        apply_fourier_encoding=True,
        max_freq=10,
        num_freq_bands=6,
        constant_mapping=False,
        max_position=1600,
        num_layers=4,
        num_latents=16,
        latent_dim=32,
        cross_num_heads=4,
        cross_head_dim=32,
        self_attend_heads=4,
        self_attend_dim=32,
        transformer_depth=2,
        attention_dropout=0.,
        ff_hidden_dim=64,
        ff_dropout=0.,
        activation=None,
    ).cuda()

    model = nn.Sequential(
        Rearrange("b c h w -> b h w c"),
        model,
    )

    dataset_train, dataset_test = get_cifar10_dataset()
    train_model_classifier(model, dataset_train, dataset_test, batch_size=128)


if __name__ == "__main__":
    main()



from copy import deepcopy
from typing import Optional, Callable, List

import torch
import torchmetrics
from colorama import Fore
from torch import nn
from torch.nn import functional as F, CrossEntropyLoss
from torchmetrics import Accuracy, AUROC, AveragePrecision, Dice, F1Score
from tqdm import tqdm


def train_model_classifier(
        model: nn.Module,
        train_dataset: torch.utils.data.Dataset,
        test_dataset: torch.utils.data.Dataset,
        loss_fn: Optional[Callable] = None,
        optimizer: Optional[torch.optim.Optimizer] = None,
        num_epochs: int = 100,
        batch_size: int = 64,
        print_every_x_steps: int = 1,
        metrics: Optional[List[torchmetrics.Metric]] = None
):
    num_classes = len(train_dataset.classes)

    # init loss function
    if loss_fn is None:
        loss_fn = CrossEntropyLoss()

    if metrics is None:
        metrics = [
            Accuracy(num_classes=num_classes).cuda(),
            AveragePrecision(num_classes=num_classes).cuda(),
            AUROC(num_classes=num_classes).cuda(),
            Dice(num_classes=num_classes).cuda(),
            F1Score(num_classes=num_classes).cuda()
        ]

    train_model(model, train_dataset, test_dataset, loss_fn, optimizer, num_epochs, batch_size, print_every_x_steps, metrics)


def train_model(
        model: nn.Module,
        train_dataset: torch.utils.data.Dataset,
        test_dataset: torch.utils.data.Dataset,
        loss_fn: Callable,
        optimizer: Optional[torch.optim.Optimizer] = None,
        num_epochs: int = 100,
        batch_size: int = 64,
        print_every_x_steps: int = 1,
        metrics: Optional[List[torchmetrics.Metric]] = None
):
    # init optimizer
    if optimizer is None:
        optimizer = torch.optim.Adam(model.parameters(), lr=0.001)

    # init loss function
    assert loss_fn is not None, "loss_fn must not be None"

    if metrics is None:
        metrics = []

    # get train loader
    train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
    train_metrics = metrics

    # get test loader
    test_loader = None
    test_metrics = None
    if test_dataset is not None:
        test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=batch_size, shuffle=True)
        test_metrics = deepcopy(metrics)

    # train model
    for epoch in range(num_epochs):
        __train_epoch(epoch, model, optimizer, print_every_x_steps, loss_fn, train_metrics, train_loader)

        if test_loader is not None:
            __test_epoch(epoch, model, print_every_x_steps, loss_fn, test_metrics, test_loader)


def __epoch(epoch, model, optimizer, print_every_x_steps, loss_fn, metrics, loader, color, mode):
    progress_bar = tqdm(loader)
    for i, (data, target) in enumerate(progress_bar):
        # Forward
        data, target = data.cuda(), target.cuda()
        output = model(data)

        # Compute loss
        loss = loss_fn(output, target)

        # Apply optimizer
        if optimizer is not None:
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

        # Update metrics
        for metric in metrics:
            metric.update(output, target)

        # Print progress
        if i % print_every_x_steps == 0:
            # prepare metric log
            metric_log = ''
            for metric in metrics:
                metric_log += f'{metric.__repr__()[:-2]}: {metric.compute():.4f} | '

            loss_name = "Loss" if len(loss_fn.__repr__()) > 30 else loss_fn.__repr__()[:-2]
            loss_log = f'{loss_name}: {loss.item():.4f} '

            description = color + f"{mode} Epoch: {epoch}, Step: {i} | {loss_log} | {metric_log}"
            progress_bar.set_description_str(description, refresh=False)


def __train_epoch(epoch, model, optimizer, print_every_x_steps, loss_fn, metrics, train_loader):
    __epoch(epoch, model, optimizer, print_every_x_steps, loss_fn, metrics, train_loader, Fore.CYAN, 'Train')


@torch.no_grad()
def __test_epoch(epoch, model, print_every_x_steps, loss_fn, metrics, test_loader):
    __epoch(epoch, model, None, print_every_x_steps, loss_fn, metrics, test_loader, Fore.YELLOW, 'Test')



from typing import Optional

from torch import nn


class DoubleConvBlock(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            mid_channels: Optional[int] = None,
            kernel_size: int = 3,
            padding: int = 1,
    ):
        super().__init__()

        if mid_channels is None:
            mid_channels = out_channels

        self.net = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=kernel_size, padding=padding),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(),
            nn.Conv2d(mid_channels, out_channels, kernel_size=kernel_size, padding=padding),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
        )

    def forward(self, x):
        return self.net(x)



from typing import Callable, Optional

import torch
from torch import nn

from vision_models_playground.components.convolutions.conv_attend import ConvAttend
from vision_models_playground.components.convolutions.conv_embedding import ConvEmbedding


class ConvTransformer(nn.Module):

    def __init__(
            self,
            in_channels: int = 3,
            patch_size: int = 4,
            patch_stride: int = 4,
            patch_padding: int = 0,
            embedding_dim: int = 512,
            depth: int = 4,
            num_heads: int = 8,
            ff_hidden_dim: int = 2048,
            qkv_bias: bool = False,
            drop_rate: float = 0.0,
            attn_drop_rate: float = 0.0,
            drop_path_rate: float = 0.0,
            activation: Optional[Callable] = None,
            **kwargs
    ):
        super().__init__()

        self.embedding_dim = embedding_dim
        self.num_features = embedding_dim

        self.patch_embedding = ConvEmbedding(
            patch_size=patch_size,
            in_channels=in_channels,
            stride=patch_stride,
            padding=patch_padding,
            out_channels=embedding_dim,
            apply_norm=True
        )

        self.positional_drop = nn.Dropout2d(drop_rate)
        self.drop_path_rate = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]

        self.layers = nn.ModuleList([])

        for i in range(depth):
            self.layers.append(
                ConvAttend(
                    in_channels=embedding_dim,
                    out_channels=embedding_dim,
                    num_heads=num_heads,
                    dim_per_head=embedding_dim // num_heads,
                    qkv_bias=qkv_bias,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=self.drop_path_rate[i],
                    activation=activation,
                    ff_hidden_dim=ff_hidden_dim,
                    output_2d=True,
                    **kwargs
                )
            )

    def forward(self, x):
        x = self.patch_embedding(x)
        x = self.positional_drop(x)

        for layer in self.layers:
            x = layer(x)

        return x


def main():
    conv_transformer = ConvTransformer(
        patch_size=4,
        patch_stride=4,
        patch_padding=0,
        in_channels=3,
        embedding_dim=32,
        num_heads=8,
        ff_hidden_dim=256,
    )

    x = torch.randn(1, 3, 64, 64)
    y = conv_transformer(x)
    print(y.shape)


if __name__ == "__main__":
    main()



from typing import Optional, Callable, Literal, List, Union

import torch
from einops import repeat
from torch import nn

from vision_models_playground.components.attention.transformer_decoder_layer import TransformerDecoderLayer
from vision_models_playground.utility.masks import create_triangular_mask


class TransformerDecoder(nn.Module):
    def __init__(
            self,
            target_dim: int,
            context_dim: int,
            depth: int,
            heads: int,
            head_dim: int,
            mlp_dim: int,
            mlp_dropout: float = 0.0,
            attention_dropout: float = 0.0,
            apply_rotary_emb: bool = False,
            activation: Optional[Callable] = None,
            drop_path: float = 0.0,
            norm_type: Literal['pre_norm', 'post_norm'] = 'pre_norm',
    ):
        """
        Args:
            target_dim:
                Dimension of the input tensor.

            context_dim:
                Dimension of the input tensor.

            depth:
                Number of layers in the encoder.

            heads:
                Number of attention heads.

            head_dim:
                Dimension of each attention head.

            mlp_dim:
                Dimension of the MLP.

            mlp_dropout:
                Dropout probability for the MLP.

            attention_dropout:
                Dropout probability for the attention.

            apply_rotary_emb:
                Whether to apply rotary embedding to the queries.

            activation:
                Activation function for the MLP.

            drop_path:
                Dropout probability for the drop path.

            norm_type:
                Type of normalization to use.
        """

        super().__init__()

        def get_transformer_decoder_layer() -> nn.Module:
            return TransformerDecoderLayer(
                target_dim=target_dim,
                context_dim=context_dim,
                heads=heads,
                head_dim=head_dim,
                mlp_dim=mlp_dim,
                mlp_dropout=mlp_dropout,
                attention_dropout=attention_dropout,
                apply_rotary_emb=apply_rotary_emb,
                activation=activation,
                drop_path=drop_path,
                norm_type=norm_type,
            )

        self.layers = nn.ModuleList([])

        for _ in range(depth):
            self.layers.append(get_transformer_decoder_layer())

    def forward(
            self,
            target: torch.Tensor,
            context: Union[torch.Tensor, List[torch.Tensor]],
            target_mask: Optional[torch.Tensor] = None,
            context_mask: Optional[torch.Tensor] = None,
            causal: bool = False,
    ) -> torch.Tensor:
        """
        Args:
            target:
                Target tensor. [batch_size, seq_len_1, dim_1]

            context:
                Context tensor. [batch_size, seq_len_2, dim_2]
                or a list [batch_size, seq_len_i, dim_i], where len(context) = depth.

            target_mask:
                Mask for the target tensor of shape [batch_size, seq_len_1]
                or [batch_size, seq_len_1, seq_len_1].

            context_mask:
                Mask for context tensor of shape [batch_size, seq_len_2]
                or [batch_size, seq_len_1, seq_len_2].

            causal:
                Whether to use causal attention to the target.
        """

        # Assert default values.
        context_list = context
        if isinstance(context, torch.Tensor):
            context_list = [context for _ in range(len(self.layers))]

        # Make sure the context_list was correctly provided.
        assert len(context_list) == len(self.layers)

        # Compute the mask for the target.
        target_mask = self.compute_2d_mask(target_mask)
        target_mask = self.compute_causal_mask(target_mask, causal)

        # Compute the mask for the context.
        context_mask = self.compute_2d_mask(context_mask, target.shape[1])

        # Forward pass.
        for transformer_decoder_layer, context in zip(self.layers, context_list):
            target = transformer_decoder_layer(target, context, target_mask, context_mask)

        return target

    @staticmethod
    def compute_causal_mask(target_mask, causal):
        if causal and target_mask is not None:
            triangular_mask = create_triangular_mask(target_mask.shape[0], target_mask.shape[1], device=target_mask.device)
            target_mask = target_mask & triangular_mask

        return target_mask

    @staticmethod
    def compute_2d_mask(mask: torch.Tensor, target_len: Optional[int] = None) -> torch.Tensor:
        if mask is not None and mask.dim() == 2:
            if target_len is None:
                target_len = mask.size(1)
            mask = repeat(mask, "b n -> b l n", l=target_len)

        return mask


def main():
    decoder = TransformerDecoder(
        target_dim=64,
        context_dim=128,
        depth=6,
        heads=8,
        head_dim=64,
        mlp_dim=2048,
        mlp_dropout=0.1,
        attention_dropout=0.1,
        apply_rotary_emb=True,
        activation=nn.ReLU(),
        drop_path=0.1,
        norm_type='pre_norm',
    )

    target = torch.randn(2, 6, 64)
    context = torch.randn(2, 10, 128)

    target_mask = torch.ones(2, 6).bool()
    target_mask[:, 3:] = 0
    context_mask = torch.ones(2, 10).bool()
    context_mask[:, 8:] = 0

    output = decoder(target, context, target_mask, context_mask, causal=True)
    print(output.shape)


if __name__ == '__main__':
    main()



from typing import Optional, Callable, Literal, List, Union

import torch
from einops import repeat
from torch import nn

from vision_models_playground.components.attention.transformer_encoder_layer import TransformerEncoderLayer


class TransformerEncoder(nn.Module):
    def __init__(
            self,
            dim: int,
            depth: int,
            heads: int,
            head_dim: int,
            mlp_dim: int,
            mlp_dropout: float = 0.0,
            attention_dropout: float = 0.0,
            apply_rotary_emb: bool = False,
            activation: Optional[Callable] = None,
            drop_path: float = 0.0,
            norm_type: Literal['pre_norm', 'post_norm'] = 'pre_norm',
    ):
        """
        Args:
            dim:
                Dimension of the input tensor.

            depth:
                Number of layers in the encoder.

            heads:
                Number of attention heads.

            head_dim:
                Dimension of each attention head.

            mlp_dim:
                Dimension of the MLP.

            mlp_dropout:
                Dropout probability for the MLP.

            attention_dropout:
                Dropout probability for the attention.

            apply_rotary_emb:
                Whether to apply rotary embedding to the queries.

            activation:
                Activation function for the MLP.

            drop_path:
                Dropout probability for the drop path.

            norm_type:
                Type of normalization to use.
            """

        super().__init__()

        def get_transformer_encoder_layer() -> nn.Module:
            return TransformerEncoderLayer(
                dim=dim,
                heads=heads,
                head_dim=head_dim,
                mlp_dim=mlp_dim,
                mlp_dropout=mlp_dropout,
                attention_dropout=attention_dropout,
                apply_rotary_emb=apply_rotary_emb,
                activation=activation,
                drop_path=drop_path,
                norm_type=norm_type,
            )

        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(get_transformer_encoder_layer())

    def forward(
            self,
            x: torch.Tensor,
            mask: Optional[torch.Tensor] = None,
    ) -> Union[torch.Tensor, List[torch.Tensor]]:
        """
        Args:
            x:
                Input tensor of shape [batch_size, seq_len, dim].

            mask:
                Mask tensor of shape [batch_size, seq_len] or [batch_size, seq_len, seq_len].

        Returns:
            Output tensor of shape [batch_size, seq_len, dim]
        """

        # Compute the mask for the input.
        mask = self.compute_2d_mask(mask)

        # Forward pass
        for transformer_encoder_layer in self.layers:
            x = transformer_encoder_layer(x=x, mask=mask)

        return x

    @staticmethod
    def compute_2d_mask(mask: torch.Tensor, target_len: Optional[int] = None) -> torch.Tensor:
        if mask is not None and mask.dim() == 2:
            if target_len is None:
                target_len = mask.size(1)
            mask = repeat(mask, "b n -> b l n", l=target_len)

        return mask


def main():
    encoder = TransformerEncoder(
        dim=512,
        depth=6,
        heads=8,
        head_dim=64,
        mlp_dim=512,
        mlp_dropout=0.1,
        attention_dropout=0.1,
        apply_rotary_emb=True,
        activation=nn.ReLU(),
        norm_type='post_norm',
    )

    x = torch.randn(4, 10, 512)
    mask = torch.ones(4, 10).bool()
    mask[:, 5:] = 0

    print(encoder(x, mask).shape)


if __name__ == '__main__':
    main()


from typing import List

from einops import rearrange
from torch import nn
from torch.nn import functional as F

from vision_models_playground.utility.datasets import get_mnist_dataset, to_autoencoder_dataset
from vision_models_playground.utility.train_models import train_model


class FeedForwardAutoencoder(nn.Module):
    def __init__(
            self,
            input_dim: int,
            hidden_dims: List[int],
    ):

        super(FeedForwardAutoencoder, self).__init__()

        if hidden_dims is None:
            hidden_dims = [256, 128]

        net_dims = [input_dim] + hidden_dims

        encoder_layers = nn.ModuleList([])
        for i in range(len(net_dims) - 1):
            encoder_layers.append(nn.Linear(net_dims[i], net_dims[i + 1]))
            encoder_layers.append(nn.ReLU())

        self.encoder = nn.Sequential(*encoder_layers)

        decoder_layers = nn.ModuleList([])
        for i in range(len(net_dims) - 1, 0, -1):
            decoder_layers.append(nn.Linear(net_dims[i], net_dims[i - 1]))
            if i != 1:
                continue
            decoder_layers.append(nn.ReLU())

        self.decoder = nn.Sequential(*decoder_layers)

    def forward(self, x):
        x = x.float()
        initial_shape = x.shape
        x = rearrange(x, "b ... -> b (...)")
        x = self.encoder(x)
        x = self.decoder(x)
        x = x.reshape(initial_shape)

        return x

    def decode(self, x):
        return self.decoder(x)

    def encode(self, x):
        return self.encoder(x)


def main():
    model = FeedForwardAutoencoder(
        input_dim=28 * 28 * 1,
        hidden_dims=[256, 128],
    ).cuda()
    train_dataset, test_dataset = get_mnist_dataset()
    train_dataset, test_dataset = map(lambda d: to_autoencoder_dataset(d), [train_dataset, test_dataset])

    # use train_dataset to train the model
    train_model(model, train_dataset, test_dataset, num_epochs=100, loss_fn=F.mse_loss)


if __name__ == '__main__':
    main()


from typing import List

import torch
from einops import rearrange
from torch import nn

from vision_models_playground.utility.datasets import get_cifar10_dataset
from vision_models_playground.utility.train_models import train_model_classifier


class FeedForwardClassifier(nn.Module):
    def __init__(
            self,
            input_dim: int = 28 * 28,
            hidden_dims: List[int] = None,
            num_classes: int = 10,
    ):
        super(FeedForwardClassifier, self).__init__()

        if hidden_dims is None:
            hidden_dims = [256, 128]

        net_dims = [input_dim] + hidden_dims + [num_classes]

        layers = []
        for i in range(len(net_dims) - 1):
            layers.append(nn.Linear(net_dims[i], net_dims[i + 1]))
            if i != len(net_dims) - 2:
                layers.append(nn.ReLU())
            else:
                layers.append(nn.Softmax(dim=1))

        self.net = nn.Sequential(*layers)

    def forward(self, x):
        x = rearrange(x, "b ... -> b (...)")
        return self.net(x)

    def predict(self, x):
        with torch.no_grad():
            return self.forward(x).argmax(dim=1)

    def predict_prob(self, x):
        with torch.no_grad():
            return self.forward(x)


def main():
    model = FeedForwardClassifier(
        input_dim=32 * 32 * 3,
        hidden_dims=[256, 128],
        num_classes=10,
    ).cuda()
    train_dataset, test_dataset = get_cifar10_dataset()
    train_model_classifier(model, train_dataset, test_dataset, num_epochs=100)


if __name__ == '__main__':
    main()



import torch
from einops import rearrange
from torch import nn
from torch.nn import functional as F

from vision_models_playground.utility.datasets import get_cifar10_dataset
from vision_models_playground.utility.train_models import train_model_classifier


class ConvolutionalClassifier(nn.Module):
    def __init__(
            self,
            num_classes: int = 10,
            channels: int = 1
    ):
        # create a sequential model that has convolutions and batch normalization
        super(ConvolutionalClassifier, self).__init__()
        self.conv1 = nn.Conv2d(channels, 32, 3, padding=1)
        self.conv2 = nn.Conv2d(32, 64, 3, padding=1)

        # create batch normalizations
        self.bn1 = nn.BatchNorm2d(32)
        self.bn2 = nn.BatchNorm2d(64)

        # create two linear components
        self.fc1 = nn.LazyLinear(256)
        self.fc2 = nn.Linear(256, num_classes)

        # create a dropout layer
        self.dropout = nn.Dropout(0.2)

    def forward(self, x):
        # pass the input through the convolutional components
        x = self.conv1(x)
        x = self.bn1(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2, 2)

        x = self.conv2(x)
        x = self.bn2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2, 2)

        # pass the input through the linear layer
        x = rearrange(x, "b ... -> b (...)")
        x = self.dropout(x)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = F.softmax(x, dim=1)

        return x

    def predict(self, x):
        return self.forward(x).argmax(dim=1)

    def predict_prob(self, x):
        with torch.no_grad():
            return self.forward(x)


def main():
    model = ConvolutionalClassifier(channels=3).cuda()
    train_dataset, test_dataset = get_cifar10_dataset()
    train_model_classifier(model, train_dataset, test_dataset, num_epochs=100)


if __name__ == '__main__':
    main()



from torch import nn


class ConvTransposedBlock(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            stride: int = 1,
            apply_relu: bool = True,
            kernel_size: int = 3,
            padding: int = 1,
            bias: bool = False
    ):
        super(ConvTransposedBlock, self).__init__()

        self.sequential = nn.Sequential(
            nn.ConvTranspose2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                bias=bias
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU() if apply_relu else nn.Identity()
        )

    def forward(self, x):
        out = self.sequential(x)
        return out



import torch
from torch import nn
from torch.nn import functional as F

from vision_models_playground.components.convolutions.conv_block import ConvBlock


class ResidualBlock(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            stride: int = 1
    ):
        super(ResidualBlock, self).__init__()
        self.conv1 = ConvBlock(in_channels, out_channels, stride=stride)
        self.conv2 = ConvBlock(out_channels, out_channels, stride=1, apply_relu=False)

        self.identity_reshape = ConvBlock(
            in_channels, out_channels, stride=stride, apply_relu=False, kernel_size=1, padding=0
        ) if in_channels != out_channels else nn.Identity()

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.conv2(out)

        out += self.identity_reshape(identity)
        out = F.relu(out)

        return out


def main():
    block = ResidualBlock(in_channels=3, out_channels=128, stride=1)

    x = torch.randn(1, 3, 300, 300)
    out = block(x)
    print(out.shape)


if __name__ == "__main__":
    main()



import torch
from torch import nn


class DropPath(nn.Module):
    """
    Drop paths (entire batches) (Stochastic Depth) per sample (when applied in main path of residual blocks).
    """

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        """
        :param drop_prob: probability of dropping a path
        :param scale_by_keep: scale by the number of paths to drop
        """

        assert 0.0 <= drop_prob <= 1.0, "drop_prob must be between 0.0 and 1.0"

        super(DropPath, self).__init__()

        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: input tensor
        :return: output tensor
        """

        # Check if the dropout must happen
        if self.drop_prob == 0.0 or not self.training:
            return x

        # Get a tensor resembling the number of paths to drop
        keep_prob = 1.0 - self.drop_prob
        batch_size = x.shape[0]
        num_dim = x.dim()
        shape = [batch_size] + [1] * (num_dim - 1)  # create [B, 1, 1, ..., 1]
        random_tensor = torch.zeros(shape, dtype=x.dtype).to(x.device).bernoulli_(keep_prob)

        # Scale the not dropped paths by the average number of paths dropped
        if self.scale_by_keep:
            random_tensor = random_tensor / keep_prob

        # Apply the dropout
        return x * random_tensor


def main():
    drop_path = DropPath(drop_prob=0.5, scale_by_keep=True)

    x = torch.randn(4, 2, 3, 3)
    out = drop_path(x)


if __name__ == "__main__":
    main()



from typing import Optional

import torch
from torch import nn

from vision_models_playground.components.attention.attend import Attend


class Compressor(nn.Module):
    """
    Compression that uses attention mechanism with information from input
    """

    def __init__(
            self,
            input_dim: int,
            output_len: int,
            output_dim: Optional[int] = None,
            attn_heads: int = 8,
            attn_dim_head: int = 64,
            attn_dropout: float = 0.0,
            hidden_dim: int = 256,
            ff_dropout: float = 0.0
    ):
        """
        Args:
            input_dim: Dimension of the input.
            output_len: Length of the output.
            output_dim: Dimension of the output.
            attn_heads: Number of attention heads.
            attn_dim_head: Dimension of the attention heads.
            attn_dropout: Dropout probability for the attention.
            hidden_dim: Dimension of the hidden components.
            ff_dropout: Dropout probability for the feed-forward components.
        """

        super(Compressor, self).__init__()

        if output_dim is None:
            output_dim = input_dim

        self.cross_attend = Attend(
            query_dim=input_dim,
            context_dim=input_dim,
            num_heads=attn_heads,
            head_dim=attn_dim_head,
            attention_dropout=attn_dropout,
            apply_rotary_emb=False,
            ff_hidden_dim=hidden_dim,
            ff_dropout=ff_dropout
        )

        self.channel_reshape = nn.Sequential(
            nn.Linear(input_dim, output_dim, bias=False),
        ) if output_dim != input_dim else nn.Identity()

        self.output_len = output_len

    def forward(self, x):
        """
        Args:
            x: Input tensor of shape (batch_size, seq_len, dim).
        Returns:
            Output tensor of shape (batch_size, output_len, dim).
        """

        assert x.shape[1] >= self.output_len, "Input sequence length must be >= output length"

        queries = x[:, :self.output_len, :]
        compressed = self.cross_attend(queries=queries, context=x)
        compressed = self.channel_reshape(compressed)

        return compressed


def main():
    x = torch.randn(1, 64, 16)

    compressor = Compressor(
        input_dim=16,
        output_len=32,
    )

    out = compressor(x)
    print(out.shape)


if __name__ == "__main__":
    main()



# Classes
from vision_models_playground.components.convolutions.downscale_block import DownscaleBlock
from vision_models_playground.components.convolutions.conv_embedding import ConvEmbedding
from vision_models_playground.components.convolutions.conv_block import ConvBlock
from vision_models_playground.components.convolutions.residual_block import ResidualBlock
from vision_models_playground.components.convolutions.upscale_block import UpscaleBlock
from vision_models_playground.components.convolutions.double_conv_block import DoubleConvBlock
from vision_models_playground.components.convolutions.conv_attend import ConvAttend
from vision_models_playground.components.convolutions.conv_attention import ConvAttention
from vision_models_playground.components.convolutions.conv_transposed_block import ConvTransposedBlock
from vision_models_playground.components.convolutions.upscale_concat_block import UpscaleConcatBlock
from vision_models_playground.components.convolutions.conv_transformer import ConvTransformer
from vision_models_playground.components.convolutions.conv_projection import ConvProjection
from vision_models_playground.components.convolutions.bottleneck_block import BottleneckBlock



import torch
from torch import nn
from torch.nn import functional as F


class Mish(nn.Module):
    """
    Swish - Described in: https://arxiv.org/abs/1908.08681
    """
    def forward(self, x):
        return x * F.softplus(x).tanh()


def main():
    import matplotlib.pyplot as plt

    x = torch.linspace(-5, 5, 100)
    y = Mish()(x)

    plt.plot(x.numpy(), y.numpy(), label="Swish")
    plt.legend()
    plt.show()


if __name__ == "__main__":
    main()


from typing import Literal

import torch
from torch import nn

from vision_models_playground.components.convolutions.double_conv_block import DoubleConvBlock


class DownscaleBlock(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            scale: int = 2,
            pooling_type: Literal["avg", "max"] = "max",
            conv_kernel_size: int = 3,
            conv_padding: int = 1,
    ):
        super().__init__()

        pool_class = nn.MaxPool2d if pooling_type == "max" else nn.AvgPool2d

        self.net = nn.Sequential(
            pool_class(kernel_size=scale),
            DoubleConvBlock(in_channels, out_channels, kernel_size=conv_kernel_size, padding=conv_padding),
        )

    def forward(self, x):
        return self.net(x)


def main():
    block = DownscaleBlock(in_channels=3, out_channels=64, scale=2, pooling_type="max", conv_kernel_size=3, conv_padding=1)
    x = torch.randn(1, 3, 32, 32)
    print(block(x).shape)


if __name__ == "__main__":
    main()


from typing import Callable, Optional

import torch
from torch import nn

from vision_models_playground.utility.functions import exists


class PreNorm(nn.Module):
    def __init__(self, dim: int, fn: Callable, context_dim: Optional[int] = None):
        super().__init__()
        self.fn = fn
        self.norm_queries = nn.LayerNorm(dim)
        self.norm_context = nn.LayerNorm(context_dim) if exists(context_dim) else None

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        x = self.norm_queries(x)

        if exists(self.norm_context):
            context = kwargs['context']
            normed_context = self.norm_context(context)
            kwargs.update(context=normed_context)

        return self.fn(x, **kwargs)



from typing import Literal

import torch
from torch import nn

from einops import repeat
from einops.layers.torch import Rearrange

from vision_models_playground.components.attention import TransformerEncoder
from vision_models_playground.utility.datasets import get_cifar10_dataset
from vision_models_playground.utility.train_models import train_model_classifier


class VisionTransformer(nn.Module):
    def __init__(
            self,
            *,
            image_size: int,
            patch_size: int,
            num_classes: int,
            projection_dim: int,
            depth: int,
            heads: int,
            mlp_dim: int,
            pool: Literal["cls", "mean"] = 'mean',
            dim_head: int = 64,
            dropout: float = 0.0,
            emb_dropout: float = 0.0,
            apply_rotary_emb: bool = False
    ):
        super().__init__()
        image_height, image_width = image_size, image_size
        patch_height, patch_width = patch_size, patch_size

        assert image_height % patch_height == 0 and image_width % patch_width == 0, \
            'Image dimensions must be divisible by the patch size.'

        num_patches = (image_height // patch_height) * (image_width // patch_width)

        self.to_patch_embedding = nn.Sequential(
            Rearrange('b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1=patch_height, p2=patch_width),
            nn.LazyLinear(projection_dim),
        )

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, projection_dim))
        self.cls_token = nn.Parameter(torch.randn(1, 1, projection_dim))
        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = TransformerEncoder(
            dim=projection_dim,
            depth=depth,
            heads=heads,
            head_dim=dim_head,
            mlp_dim=mlp_dim,
            mlp_dropout=dropout,
            attention_dropout=dropout,
            apply_rotary_emb=apply_rotary_emb
        )

        self.pool = pool

        self.mlp_head = nn.Sequential(
            nn.LayerNorm(projection_dim),
            nn.Linear(projection_dim, num_classes)
        )

    def forward(self, img: torch.Tensor):
        x = self.to_patch_embedding(img)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, '1 n d -> b n d', b=b)

        x = torch.cat([cls_tokens, x], dim=1)
        x += self.pos_embedding[:, :(n + 1)]
        x = self.dropout(x)

        x = self.transformer(x)

        x = x.mean(dim=1) if self.pool == 'mean' else x[:, 0]

        return self.mlp_head(x)


def main():
    model = VisionTransformer(
        image_size=32,
        patch_size=8,
        num_classes=10,
        projection_dim=256,
        depth=6,
        heads=8,
        mlp_dim=512,
        dropout=0.2,
        emb_dropout=0.2,
        apply_rotary_emb=True,
        pool="cls",
    ).cuda()
    train_dataset, test_dataset = get_cifar10_dataset()
    train_model_classifier(model, train_dataset, test_dataset, num_epochs=100)


if __name__ == '__main__':
    main()



# Classes
from vision_models_playground.components.attention.transformer_encoder_layer import TransformerEncoderLayer
from vision_models_playground.components.attention.attention import Attention
from vision_models_playground.components.attention.transformer_decoder_layer import TransformerDecoderLayer
from vision_models_playground.components.attention.pre_norm import PreNorm
from vision_models_playground.components.attention.post_norm import PostNorm
from vision_models_playground.components.attention.attend import Attend
from vision_models_playground.components.attention.feed_forward import FeedForward
from vision_models_playground.components.attention.transformer_decoder import TransformerDecoder
from vision_models_playground.components.attention.compressor import Compressor
from vision_models_playground.components.attention.transformer import Transformer
from vision_models_playground.components.attention.transformer_encoder import TransformerEncoder



from typing import Optional, Callable

import torch
from torch import nn

from vision_models_playground.components.attention.attention import Attention
from vision_models_playground.components.attention.feed_forward import FeedForward
from vision_models_playground.components.attention.pre_norm import PreNorm
from vision_models_playground.components.dropout import DropPath


class Attend(nn.Module):
    def __init__(
            self,
            query_dim: int,
            context_dim: Optional[int] = None,
            num_heads: int = 8,
            head_dim: int = 64,
            attention_dropout: float = 0.0,
            apply_rotary_emb: bool = False,
            ff_hidden_dim: int = 256,
            ff_dropout: float = 0.0,
            activation: Callable = None,
            drop_path: float = 0.0,
    ):
        super(Attend, self).__init__()

        self.attention = PreNorm(
            dim=query_dim,
            fn=Attention(
                query_dim=query_dim,
                context_dim=context_dim,
                heads=num_heads,
                dim_head=head_dim,
                dropout=attention_dropout,
                apply_rotary_emb=apply_rotary_emb
            ),
            context_dim=context_dim
        )

        self.mlp = PreNorm(
            dim=query_dim,
            fn=FeedForward(
                dim=query_dim,
                hidden_dim=ff_hidden_dim,
                dropout=ff_dropout,
                activation=activation
            )
        )

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(
            self,
            queries: torch.Tensor,
            context: Optional[torch.Tensor] = None,
            mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:

        x = queries + self.drop_path(self.attention(x=queries, context=context, mask=mask))
        x = x + self.drop_path(self.mlp(x=x))

        return x



import torch
from einops import rearrange
from torch.utils.data import TensorDataset
from torchvision import datasets, transforms


def to_autoencoder_dataset(dataset: torch.utils.data.Dataset):
    images = dataset.data

    if images.dtype == torch.uint8:
        images = images.float() / 255.0

    return TensorDataset(images, images)


def get_mnist_dataset(root: str='./data'):
    mnist_train = datasets.MNIST(root=root, train=True, download=True, transform=transforms.ToTensor())
    mnist_test = datasets.MNIST(root=root, train=False, download=True, transform=transforms.ToTensor())

    return mnist_train, mnist_test


def get_cifar10_dataset(root: str='./data'):
    cifar10_train = datasets.CIFAR10(root=root, train=True, download=True, transform=transforms.ToTensor())
    cifar10_test = datasets.CIFAR10(root=root, train=False, download=True, transform=transforms.ToTensor())

    return cifar10_train, cifar10_test


def get_tourism_dataset(root: str='./data', name: str='notredame'):
    photo_tour_train = datasets.PhotoTour(root=root, name=name, train=True, download=True)
    photo_tour_test = datasets.PhotoTour(root=root, name=name, train=False, download=True)

    return photo_tour_train, photo_tour_test


def get_n_images(dataset: torch.utils.data.Dataset, num_images: int):
    train_loader = torch.utils.data.DataLoader(
        dataset,
        batch_size=1,
        shuffle=True
    )

    images = []
    for i, x in enumerate(train_loader):
        if i == num_images:
            break

        if isinstance(x, tuple):
            x = x[0]

        if len(x.shape) == 3:
            # Add batch_size dimension
            x = rearrange(x, "... -> 1 ...")

        images.append(x)

    images = torch.cat(images, dim=0)
    return images



from typing import Literal

import torch
from torch import nn
from torch.nn import functional as F

from vision_models_playground.components.convolutions.double_conv_block import DoubleConvBlock


class UpscaleConcatBlock(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            scale: int = 2,
            conv_kernel_size: int = 3,
            conv_padding: int = 1,
            method: Literal["nearest", "linear", "bilinear", "bicubic", "trilinear", "conv"] = "conv",
            crop: bool = False
    ):
        super().__init__()

        self.upscale = nn.Sequential(
            nn.Upsample(scale_factor=scale, mode=method, align_corners=True),
            nn.Conv2d(in_channels, in_channels // 2, kernel_size=scale+1, padding=scale-1)
        ) if method != "conv" else \
            nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=scale, stride=scale)

        self.conv = DoubleConvBlock(in_channels, out_channels, kernel_size=conv_kernel_size, padding=conv_padding)

        self.crop = crop

    def forward(self, x1, x2):
        # Upscale x1
        x1 = self.upscale(x1)

        # pad the first image to match the second shape
        diff_y = x2.shape[2] - x1.shape[2]
        left_y = diff_y // 2
        right_y = diff_y - left_y

        diff_x = x2.shape[3] - x1.shape[3]
        left_x = diff_x // 2
        right_x = diff_x - left_x

        x1 = F.pad(
            x1,
            [left_y, right_y,
             left_x, right_x]
        )

        # add the two images together
        x = torch.cat([x1, x2], dim=1)

        # pass the combined image through the convolutional layer
        x = self.conv(x)

        if self.crop:
            x = x[:, :, left_y: -right_y, left_x: -right_x]

        return x


def main():
    block = UpscaleConcatBlock(in_channels=1024, out_channels=512, scale=2, conv_kernel_size=3, conv_padding=1, method="conv", crop=True)
    x1 = torch.randn(1, 1024, 28, 28)
    x2 = torch.randn(1, 512, 64, 64)
    print(block(x1, x2).shape)


if __name__ == "__main__":
    main()


from torch import nn

from vision_models_playground.components.activations import QuickGELU


class QuickGEGLU(nn.Module):
    def __init__(self):
        super(QuickGEGLU, self).__init__()
        self.quick_gelu = QuickGELU()

    def forward(self, x):
        x, gates = x.chunk(2, dim=-1)
        return x * self.quick_gelu(gates)



import torch


def create_triangular_mask(batch_size: int, length: int, device: torch.device = torch.device('cpu')):
    """
    Return a triangular mask with shape [b, n, n] where:
    b = batch_size
    n = length
    """

    b = batch_size
    n = length
    mask = torch.ones([b, n, n], dtype=torch.bool, device=device)
    mask = torch.triu(mask, diagonal=1)
    mask = ~mask
    return mask



import torch
from torch import nn
from torch.nn import functional as F


class Generator(nn.Module):
    def __init__(self, noise_dim=100, channels=1):
        """
        Initializes the Generator.
        :param noise_dim: the dimension of the noise vector
        :param channels: the number of channels in the image
        """

        super(Generator, self).__init__()

        self.noise_dim = noise_dim

        self.tconv1 = nn.ConvTranspose2d(noise_dim, 512, 4, stride=1, padding=0)
        self.bn1 = nn.BatchNorm2d(512)

        self.tconv2 = nn.ConvTranspose2d(512, 256, 4, stride=2, padding=1)
        self.bn2 = nn.BatchNorm2d(256)

        self.tconv3 = nn.ConvTranspose2d(256, 128, 4, stride=2, padding=1)
        self.bn3 = nn.BatchNorm2d(128)

        self.tconv4 = nn.ConvTranspose2d(128, channels, 4, stride=2, padding=3)

    def forward(self, x):
        """
        Generates an image from a noise vector.
        :param x: the noise vector
        :return: image
        """

        x = x.view(-1, 100, 1, 1)
        x = F.relu(self.tconv1(x))
        x = self.bn1(x)
        x = F.relu(self.tconv2(x))
        x = self.bn2(x)
        x = F.relu(self.tconv3(x))
        x = self.bn3(x)
        x = F.relu(self.tconv4(x))
        x = torch.tanh(x)

        return x



from typing import Optional, Callable, Literal

import torch
from torch import nn

from vision_models_playground.components.attention.feed_forward import FeedForward
from vision_models_playground.components.attention.attention import Attention
from vision_models_playground.components.attention.pre_norm import PreNorm
from vision_models_playground.components.attention.post_norm import PostNorm
from vision_models_playground.components.dropout import DropPath


class TransformerEncoderLayer(nn.Module):
    def __init__(
            self,
            dim: int,
            heads: int,
            head_dim: int,
            mlp_dim: int,
            mlp_dropout: float = 0.0,
            attention_dropout: float = 0.0,
            apply_rotary_emb: bool = False,
            activation: Optional[Callable] = None,
            drop_path: float = 0.0,
            norm_type: Literal['pre_norm', 'post_norm'] = 'pre_norm',
    ):
        super().__init__()

        norm_layer = PreNorm if norm_type == 'pre_norm' else PostNorm

        self.self_attention = norm_layer(
            dim=dim,
            fn=Attention(
                query_dim=dim,
                heads=heads,
                dim_head=head_dim,
                dropout=attention_dropout,
                apply_rotary_emb=apply_rotary_emb
            )
        )

        self.mlp = norm_layer(
            dim=dim,
            fn=FeedForward(
                dim=dim,
                hidden_dim=mlp_dim,
                dropout=mlp_dropout,
                activation=activation
            )
        )

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(
            self,
            x: torch.Tensor,
            mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Args:
            x: [batch_size, seq_len, dim]
            mask: [batch_size, seq_len, seq_len]
        """

        x = x + self.drop_path(self.self_attention(x=x, mask=mask))
        x = x + self.drop_path(self.mlp(x=x))

        return x



from typing import Callable

import torch
from torch import nn


class PostNorm(nn.Module):
    def __init__(self, dim: int, fn: Callable, **kwargs):
        super().__init__()
        self.fn = fn
        self.norm = nn.LayerNorm(dim)

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        return self.norm(self.fn(x, **kwargs))



from typing import Callable, List, Literal

import torch
from einops.layers.torch import Reduce, Rearrange
from torch import nn

from vision_models_playground.components.activations import QuickGELU
from vision_models_playground.components.convolutions.conv_transformer import ConvTransformer
from vision_models_playground.utility.datasets import get_cifar10_dataset
from vision_models_playground.utility.functions import get_number_of_parameters
from vision_models_playground.utility.train_models import train_model_classifier


class ConvVisionTransformer(nn.Module):
    def __init__(
            self,
            in_channels: int = 3,
            num_classes: int = 1000,
            activation: Callable = None,
            final_classifier_head: bool = True,

            patch_size: List[int] = None,
            patch_stride: List[int] = None,
            patch_padding: List[int] = None,
            embedding_dim: List[int] = None,
            depth: List[int] = None,
            num_heads: List[int] = None,
            ff_hidden_dim: List[int] = None,
            qkv_bias: List[bool] = None,
            drop_rate: List[float] = None,
            attn_drop_rate: List[float] = None,
            drop_path_rate: List[float] = None,
            kernel_size: List[int] = None,
            stride_kv: List[int] = None,
            stride_q: List[int] = None,
            padding_kv: List[int] = None,
            padding_q: List[int] = None,
            method: List[Literal['conv', 'avg', 'linear']] = None,
    ):
        super().__init__()

        if activation is None:
            activation = QuickGELU()

        # Ensure no param is None
        assert all(x is not None for x in [
            patch_size, patch_stride, patch_padding, embedding_dim, depth, num_heads, ff_hidden_dim,
            qkv_bias, drop_rate, attn_drop_rate, drop_path_rate, kernel_size, stride_kv, stride_q,
            padding_kv, padding_q, method,
        ])

        # Ensure all lists have the same length
        assert len(patch_size) == len(patch_stride) == len(patch_padding) == len(embedding_dim) == \
               len(depth) == len(num_heads) == len(ff_hidden_dim) == len(qkv_bias) == len(drop_rate) == \
               len(attn_drop_rate) == len(drop_path_rate) == len(kernel_size) == len(stride_kv) == \
               len(stride_q) == len(padding_kv) == len(padding_q) == len(method)

        self.num_transformers = len(patch_size)

        self.transformers = nn.ModuleList([])

        current_channels = in_channels
        for i in range(self.num_transformers):
            self.transformers.append(
                ConvTransformer(
                    in_channels=current_channels,
                    patch_size=patch_size[i],
                    patch_stride=patch_stride[i],
                    patch_padding=patch_padding[i],
                    embedding_dim=embedding_dim[i],
                    depth=depth[i],
                    num_heads=num_heads[i],
                    ff_hidden_dim=ff_hidden_dim[i],
                    qkv_bias=qkv_bias[i],
                    drop_rate=drop_rate[i],
                    attn_drop_rate=attn_drop_rate[i],
                    drop_path_rate=drop_path_rate[i],
                    kernel_size=kernel_size[i],
                    stride_kv=stride_kv[i],
                    stride_q=stride_q[i],
                    padding_kv=padding_kv[i],
                    padding_q=padding_q[i],
                    method=method[i],
                    activation=activation,
                )
            )

            current_channels = embedding_dim[i]

        self.head = nn.Sequential(
            Rearrange("b c h w -> b (h w) c"),
            Reduce("b l d -> b d", "mean"),
            nn.LayerNorm(current_channels),
            nn.Linear(current_channels, num_classes),
        ) if final_classifier_head else nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for transformer in self.transformers:
            x = transformer(x)

        return self.head(x)


def build_cvt_13(num_classes: int = 10, in_channels: int = 3):
    return ConvVisionTransformer(
        in_channels=in_channels,
        num_classes=num_classes,
        patch_size=[7, 3, 3],
        patch_stride=[4, 2, 2],
        patch_padding=[2, 1, 1],
        embedding_dim=[64, 192, 384],
        depth=[1, 2, 10],
        num_heads=[1, 3, 6],
        ff_hidden_dim=[256, 768, 1536],
        qkv_bias=[True, True, True],
        drop_rate=[0.0, 0.0, 0.0],
        attn_drop_rate=[0.0, 0.0, 0.0],
        drop_path_rate=[0.0, 0.0, 0.1],
        kernel_size=[3, 3, 3],
        stride_kv=[2, 2, 2],
        stride_q=[1, 1, 1],
        padding_kv=[1, 1, 1],
        padding_q=[1, 1, 1],
        method=['conv', 'conv', 'conv'],
    )


def build_cvt_21(num_classes: int = 10, in_channels: int = 3):
    return ConvVisionTransformer(
        in_channels=in_channels,
        num_classes=num_classes,
        patch_size=[7, 3, 3],
        patch_stride=[4, 2, 2],
        patch_padding=[2, 1, 1],
        embedding_dim=[64, 192, 384],
        depth=[1, 4, 16],
        num_heads=[1, 3, 6],
        ff_hidden_dim=[256, 768, 1536],
        qkv_bias=[True, True, True],
        drop_rate=[0.0, 0.0, 0.0],
        attn_drop_rate=[0.0, 0.0, 0.0],
        drop_path_rate=[0.0, 0.0, 0.1],
        kernel_size=[3, 3, 3],
        stride_kv=[2, 2, 2],
        stride_q=[1, 1, 1],
        padding_kv=[1, 1, 1],
        padding_q=[1, 1, 1],
        method=['conv', 'conv', 'conv'],
    )


def build_cvt_w24(num_classes: int = 10, in_channels: int = 3):
    return ConvVisionTransformer(
        in_channels=in_channels,
        num_classes=num_classes,
        patch_size=[7, 3, 3],
        patch_stride=[4, 2, 2],
        patch_padding=[2, 1, 1],
        embedding_dim=[64, 768, 1024],
        depth=[2, 2, 20],
        num_heads=[3, 12, 16],
        ff_hidden_dim=[256, 3072, 4096],
        qkv_bias=[True, True, True],
        drop_rate=[0.0, 0.0, 0.0],
        attn_drop_rate=[0.0, 0.0, 0.0],
        drop_path_rate=[0.0, 0.0, 0.3],
        kernel_size=[3, 3, 3],
        stride_kv=[2, 2, 2],
        stride_q=[1, 1, 1],
        padding_kv=[1, 1, 1],
        padding_q=[1, 1, 1],
        method=['conv', 'conv', 'conv'],
        final_classifier_head=True,
    )


if __name__ == '__main__':
    model = ConvVisionTransformer(
        in_channels=3,
        num_classes=10,
        patch_size=[7, 3],
        patch_stride=[4, 2],
        patch_padding=[2, 1],
        embedding_dim=[64, 192],
        depth=[2, 8],
        num_heads=[1, 3],
        ff_hidden_dim=[256, 768],
        qkv_bias=[True, True],
        drop_rate=[0.0, 0.0],
        attn_drop_rate=[0.0, 0.0],
        drop_path_rate=[0.0, 0.1],
        kernel_size=[3, 3],
        stride_kv=[2, 2],
        stride_q=[1, 1],
        padding_kv=[1, 1],
        padding_q=[1, 1],
        method=['conv', 'conv'],
    ).cuda()

    # print number of params of the model
    print(f"Number of params: {get_number_of_parameters(model) / (1024 ** 2):.3f} M")

    dataset_train, dataset_test = get_cifar10_dataset()
    train_model_classifier(model, dataset_train, dataset_test)



from typing import Optional, Callable

import torch
from einops import rearrange
from torch import nn

from vision_models_playground.components.attention import PreNorm, FeedForward
from vision_models_playground.components.convolutions.conv_attention import ConvAttention
from vision_models_playground.components.dropout import DropPath


class ConvAttend(nn.Module):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            num_heads: int,
            dim_per_head: int,
            qkv_bias: bool = False,
            drop: float = 0.0,
            attn_drop: float = 0.0,
            drop_path: float = 0.0,
            activation: Optional[Callable] = None,
            ff_hidden_dim: int = 256,
            output_2d: bool = False,
            **kwargs
    ):
        super().__init__()

        self.output_2d = output_2d

        self.norm = nn.LayerNorm(in_channels)

        self.attention = ConvAttention(
            in_channels=in_channels,
            out_channels=out_channels,
            num_heads=num_heads,
            dim_per_head=dim_per_head,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            output_drop=drop,
            **kwargs
        )

        self.mlp = PreNorm(
            dim=out_channels,
            fn=FeedForward(
                dim=out_channels,
                hidden_dim=ff_hidden_dim,
                dropout=drop,
                activation=activation,
                output_dim=out_channels
            )
        )

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.identity_reshape = nn.Identity() if in_channels == out_channels \
            else nn.Linear(in_channels, out_channels, bias=False)

    def forward(self, x):
        b, c, h, w = x.shape

        x = rearrange(x, "b c h w -> b (h w) c")
        res = x
        x = self.norm(x)
        x = rearrange(x, "b (h w) c -> b c h w", h=h)

        x = self.identity_reshape(res) + self.drop_path(self.attention(x))
        x = x + self.drop_path(self.mlp(x))

        if self.output_2d:
            x = rearrange(x, "b (h w) c -> b c h w", h=h)

        return x


def main():
    block = ConvAttend(
        in_channels=64,
        out_channels=128,
        num_heads=4,
        dim_per_head=32,
        qkv_bias=False,
        ff_hidden_dim=256,
    )

    x = torch.randn(1, 64, 32, 32)
    out = block(x)

    print(out.shape)


if __name__ == "__main__":
    main()



import vision_models_playground.components.activations
import vision_models_playground.components.attention
import vision_models_playground.components.convolutions
import vision_models_playground.components.dropout
import vision_models_playground.components.position_embedding



from typing import Optional, Callable, Literal

import torch
from torch import nn

from vision_models_playground.components.attention.feed_forward import FeedForward
from vision_models_playground.components.attention.attention import Attention
from vision_models_playground.components.attention.pre_norm import PreNorm
from vision_models_playground.components.attention.post_norm import PostNorm
from vision_models_playground.components.dropout import DropPath


class TransformerDecoderLayer(nn.Module):
    def __init__(
            self,
            target_dim: int,
            context_dim: int,
            heads: int,
            head_dim: int,
            mlp_dim: int,
            mlp_dropout: float = 0.0,
            attention_dropout: float = 0.0,
            apply_rotary_emb: bool = False,
            activation: Optional[Callable] = None,
            drop_path: float = 0.0,
            norm_type: Literal['pre_norm', 'post_norm'] = 'pre_norm',
    ):
        super().__init__()

        norm_layer = PreNorm if norm_type == 'pre_norm' else PostNorm

        self.self_attention = norm_layer(
            dim=target_dim,
            fn=Attention(
                query_dim=target_dim,
                heads=heads,
                dim_head=head_dim,
                dropout=attention_dropout,
                apply_rotary_emb=apply_rotary_emb
            )
        )

        self.cross_attention = norm_layer(
            dim=target_dim,
            fn=Attention(
                query_dim=target_dim,
                context_dim=context_dim,
                heads=heads,
                dim_head=head_dim,
                dropout=attention_dropout,
                apply_rotary_emb=apply_rotary_emb
            ),
            context_dim=context_dim,
        )

        self.mlp = norm_layer(
            dim=target_dim,
            fn=FeedForward(
                dim=target_dim,
                hidden_dim=mlp_dim,
                dropout=mlp_dropout,
                activation=activation
            )
        )

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(
            self,
            target: torch.Tensor,
            context: Optional = None,
            target_mask: Optional[torch.Tensor] = None,
            context_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Args:
            target: [batch_size, seq_len_1, dim_1]
            context: [batch_size, seq_len_2, dim_2]
            target_mask: [batch_size, seq_len_1, seq_len_1]
            context_mask: [batch_size, seq_len_1, seq_len_2]
        """

        target = target + self.drop_path(self.self_attention(x=target, mask=target_mask))
        target = target + self.drop_path(self.cross_attention(x=target, context=context, mask=context_mask))
        target = target + self.drop_path(self.mlp(x=target))

        return target



from matplotlib import pyplot as plt
from torch import nn

from torchvision.transforms import RandomHorizontalFlip, RandomPerspective, RandomResizedCrop, RandomErasing, RandomAffine, \
    RandomRotation, ColorJitter

from vision_models_playground.utility.datasets import get_mnist_dataset, get_n_images


class Augmeneter(nn.Module):
    def __init__(
            self,
            image_size: int = 32,
            background_color: float = 1.0,
            rotation_angles: int = 30
    ):
        super().__init__()

        self.net = nn.Sequential(
            RandomHorizontalFlip(),
            RandomPerspective(fill=background_color),
            RandomResizedCrop(size=image_size, scale=(0.90, 1.00)),
            RandomErasing(value=background_color, scale=(0.1, 0.15)),
            RandomAffine(degrees=rotation_angles, fill=background_color),
            RandomRotation(degrees=rotation_angles, fill=background_color),
            ColorJitter(brightness=(0.8, 1.0)),
            ColorJitter(contrast=(0.8, 1.0)),
            ColorJitter(saturation=(0.5, 1.0)),
        )

    def forward(self, x):
        return self.net(x)


def main():
    augmenter = Augmeneter(image_size=28, background_color=0.0, rotation_angles=15)
    mnist_dataset_train, _ = get_mnist_dataset()
    images = get_n_images(mnist_dataset_train, num_images=25)
    images = augmenter(images)

    plt.figure(figsize=(10, 10))
    for i in range(25):
        plt.subplot(5, 5, i + 1)
        plt.imshow(images[i].cpu().data.numpy().reshape(28, 28), cmap='gray')
        plt.axis('off')

    plt.show()


if __name__ == '__main__':
    main()


from typing import Callable

import torch
from torch import nn

from vision_models_playground.components.activations.geglu import GEGLU
from vision_models_playground.components.activations.relu_squared import ReluSquared


class FeedForward(nn.Module):
    def __init__(
            self,
            dim: int,
            hidden_dim: int,
            dropout: float = 0.0,
            activation: Callable = None,
            output_dim: int = None
    ):
        super(FeedForward, self).__init__()

        if activation is None:
            activation = ReluSquared()

        activation_dim = hidden_dim
        if isinstance(activation, GEGLU):
            assert hidden_dim % 2 == 0, "hidden_dim must be even when using GEGLU"
            activation_dim = hidden_dim // 2

        if output_dim is None:
            output_dim = dim

        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            activation,
            nn.Dropout(dropout),
            nn.Linear(activation_dim, output_dim),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x)



from typing import Literal

import torch
from torch import nn

from vision_models_playground.components.convolutions.double_conv_block import DoubleConvBlock


class UpscaleBlock(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            scale: int = 2,
            conv_kernel_size: int = 3,
            conv_padding: int = 1,
            method: Literal["nearest", "linear", "bilinear", "bicubic", "trilinear", "conv"] = "conv",
    ):
        super().__init__()

        upscale = nn.Upsample(scale_factor=scale, mode=method, align_corners=True) if method != "conv" \
            else nn.ConvTranspose2d(in_channels, in_channels, kernel_size=scale, stride=scale)

        conv = DoubleConvBlock(in_channels, out_channels, kernel_size=conv_kernel_size, padding=conv_padding)

        self.net = nn.Sequential(upscale, conv)

    def forward(self, x):
        return self.net(x)


def main():
    block = UpscaleBlock(in_channels=16, out_channels=32, scale=2, conv_kernel_size=3, conv_padding=1, mode="bicubic")
    x = torch.randn(1, 16, 32, 32)
    print(block(x).shape)


if __name__ == "__main__":
    main()



from einops import rearrange
from matplotlib import pyplot as plt


def display_images_on_grid(images, n_rows=5, n_cols=5):
    """
    Display a grid of images.
    """
    plt.figure(figsize=(10, 10))

    channels = images.shape[1]

    for i in range(n_rows * n_cols):
        plt.subplot(n_rows, n_cols, i + 1)

        image = rearrange(images[i], "c h w -> h w c")
        cmap = "gray" if channels == 1 else None

        plt.imshow(image, cmap=cmap)
        plt.axis('off')

    plt.show()


def display_image(image):
    """
    Display an image.
    """

    image = rearrange(image, "c h w -> h w c")
    channels = image.shape[0]
    cmap = "gray" if channels == 1 else None

    plt.figure(figsize=(10, 10))
    plt.imshow(image, cmap=cmap)
    plt.axis('off')
    plt.show()


from typing import Literal

import torch
from einops.layers.torch import Rearrange
from torch import nn


class ConvProjection(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int = 3,
            padding: int = 1,
            stride: int = 1,
            method: Literal['conv', 'avg', 'linear'] = 'conv',
            bias: bool = False,
    ):
        super().__init__()

        # Create network according to the method
        if method == 'conv':
            net = nn.Sequential(
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                    groups=in_channels,
                    bias=False
                ),
                nn.BatchNorm2d(in_channels),
            )

        elif method == 'avg':
            net = nn.AvgPool2d(
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                ceil_mode=True
            )

        elif method == 'linear':
            net = nn.Identity()

        else:
            raise ValueError(f'Unknown method: {method}')

        # Create projections layer
        projection = nn.Sequential(
            Rearrange("b c h w -> b (h w) c"),
            nn.Linear(in_channels, out_channels, bias=bias)
        )

        self.net = nn.Sequential(
            net, projection
        )

    def forward(self, x):
        return self.net(x)


def main():
    block = ConvProjection(in_channels=3, out_channels=60, kernel_size=3, padding=1, stride=1, method='conv')

    x = torch.randn(1, 3, 32, 32)
    out = block(x)  # [1, 32*32, 60]
    print(out.shape)


if __name__ == "__main__":
    main()


from einops import rearrange
from torch import nn


class ConvEmbedding(nn.Module):
    """
    Image to Conv Embedding
    """

    def __init__(
            self,
            patch_size: int = 7,
            in_channels: int = 3,
            out_channels: int = 64,
            stride: int = 4,
            padding: int = 2,
            apply_norm: bool = True,
    ):
        super().__init__()
        self.patch_size = patch_size

        self.proj = nn.Conv2d(
            in_channels, out_channels,
            kernel_size=patch_size,
            stride=stride,
            padding=padding
        )

        self.norm = nn.LayerNorm(out_channels) if apply_norm else None

    def forward(self, x):
        x = self.proj(x)

        b, c, h, w = x.shape

        if self.norm is not None:
            x = rearrange(x, 'b c h w -> b (h w) c')
            x = self.norm(x)
            x = rearrange(x, 'b (h w) c -> b c h w', h=h, w=w)

        return x



# Functions
from vision_models_playground.utility.datasets import to_autoencoder_dataset
from vision_models_playground.utility.datasets import get_mnist_dataset
from vision_models_playground.utility.datasets import get_cifar10_dataset
from vision_models_playground.utility.datasets import get_tourism_dataset
from vision_models_playground.utility.datasets import get_n_images
from vision_models_playground.utility.train_models import train_model_classifier
from vision_models_playground.utility.train_models import train_model
from vision_models_playground.utility.visualize import display_images_on_grid
from vision_models_playground.utility.visualize import display_image
from vision_models_playground.utility.masks import create_triangular_mask
from vision_models_playground.utility.functions import exists
from vision_models_playground.utility.functions import default
from vision_models_playground.utility.functions import get_number_of_parameters



from typing import Any

from torch import nn


def exists(val: Any) -> bool:
    return val is not None


def default(val: Any, default_val: Any) -> Any:
    return val if exists(val) else default_val


def get_number_of_parameters(model: nn.Module) -> int:
    return sum(param.numel() for param in model.parameters())



from typing import List, Optional, Type, Union

from einops import rearrange
from torch import nn

from vision_models_playground.components.convolutions.conv_block import ConvBlock
from vision_models_playground.components.convolutions.residual_block import ResidualBlock
from vision_models_playground.components.convolutions.bottleneck_block import BottleneckBlock
from vision_models_playground.utility.datasets import get_cifar10_dataset
from vision_models_playground.utility.train_models import train_model_classifier


class ResNet(nn.Module):
    def __init__(
            self,
            in_channels: int = 3,
            num_classes: int = 10,
            block: Type[Union[ResidualBlock, BottleneckBlock]] = None,
            num_layers: Optional[List[int]] = None,
            num_channels: Optional[List[int]] = None,
    ):
        super(ResNet, self).__init__()

        # assign default values to parameters
        if block is None:
            block = ResidualBlock

        if num_layers is None:
            num_layers = [3, 4, 6, 3]

        if num_channels is None:
            num_channels = [64, 128, 256, 512]

        # save parameters
        self.num_layers = num_layers
        self.num_channels = num_channels

        # build components
        self.conv1 = ConvBlock(in_channels, num_channels[0], stride=2, kernel_size=7, padding=3)
        self.max_pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layers = self.__build_layers(block)

        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.LazyLinear(num_classes)

    def __build_layers(
            self,
            block: Type[Union[ResidualBlock, BottleneckBlock]]
    ):
        layers = nn.ModuleList()
        for i in range(len(self.num_layers)):
            prev_num_channels = self.num_channels[i - 1] if i != 0 else self.num_channels[i]
            layers.append(self.__build_layer(i, prev_num_channels, block))

        return layers

    def __build_layer(
            self,
            index: int,
            prev_num_channels: int,
            block: Type[Union[ResidualBlock, BottleneckBlock]],
    ):
        layer = nn.ModuleList()
        num_layers = self.num_layers[index]
        num_channels = self.num_channels[index]

        for j in range(num_layers):
            stride = 2 if j == 0 and prev_num_channels != num_channels else 1
            input_num_channels = prev_num_channels if j == 0 else num_channels

            layer.append(block(input_num_channels, num_channels, stride=stride))

        return nn.Sequential(*layer)

    def forward(self, x):
        out = self.conv1(x)
        out = self.max_pool(out)

        for layer in self.layers:
            out = layer(out)

        out = self.avg_pool(out)
        out = rearrange(out, "b ... -> b (...)")
        out = self.fc(out)

        return out


def build_resnet_18(num_classes: int = 10, in_channels: int = 3):
    return ResNet(
        num_classes=num_classes,
        in_channels=in_channels,
        num_layers=[2, 2, 2, 2],
        num_channels=[64, 128, 256, 512],
        block=ResidualBlock
    )


def build_resnet_34(num_classes: int = 10, in_channels: int = 3):
    return ResNet(
        num_classes=num_classes,
        in_channels=in_channels,
        num_layers=[3, 4, 6, 3],
        num_channels=[64, 128, 256, 512],
        block=ResidualBlock
    )


def build_resnet_50(num_classes: int = 10, in_channels: int = 3):
    return ResNet(
        num_classes=num_classes,
        in_channels=in_channels,
        num_layers=[3, 4, 6, 3],
        num_channels=[64, 128, 256, 512],
        block=BottleneckBlock
    )


def build_resnet_101(num_classes: int = 10, in_channels: int = 3):
    return ResNet(
        num_classes=num_classes,
        in_channels=in_channels,
        num_layers=[3, 4, 23, 3],
        num_channels=[64, 128, 256, 512],
        block=BottleneckBlock
    )


def build_resnet_152(num_classes: int = 10, in_channels: int = 3):
    return ResNet(
        num_classes=num_classes,
        in_channels=in_channels,
        num_layers=[3, 8, 36, 3],
        num_channels=[64, 256, 512, 1024],
        block=BottleneckBlock
    )


def main():
    model = build_resnet_50().cuda()

    dataset_train, dataset_test = get_cifar10_dataset()
    train_model_classifier(model, dataset_train, dataset_test)


if __name__ == "__main__":
    main()



import vision_models_playground.models
import vision_models_playground.components
import vision_models_playground.utility


__version__ = "0.1.0"
__author__ = "Alexandru Stirbu"
__email__ = "Alexandru.Stirbu.Net@outlook.com"
__license__ = "MIT"
__username__ = "Akrielz"



from setuptools import setup, find_packages

setup(
    name='vision_models_playground',
    packages=find_packages(),
    version='0.1.0',
    license='MIT',
    description="Akriel's vision models playground",
    author='Alexandru Stirbu',
    author_email='stirbu.alexandru.net@outlook.com',
    url='https://github.com/Akrielz/vision_models_playground',
    keywords=[
        'artificial intelligence',
        'deep learning',
        'transformer',
        'attention mechanism',
        'computer vision',
    ],
    install_requires=[
        'einops>=0.4.1',
        'torch>=1.10.0',
        'rotary-embedding-torch>=0.1.0',
        'torchvision>=0.11.1',
        'colorama>=0.4.5',
        'tqdm>=4.64.0',
        'torchmetrics>=0.9.3',
        'numpy>=1.22.0',
        'matplotlib>=3.5.2',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
)



from typing import Literal, List

import torch
from torch import nn

from vision_models_playground.components.convolutions import UpscaleConcatBlock, DoubleConvBlock, DownscaleBlock


class UNet(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            channels: List[int],
            pooling_type: Literal["avg", "max"] = "max",
            scale: int = 2,
            conv_kernel_size: int = 3,
            conv_padding: int = 1,
            method: Literal["nearest", "linear", "bilinear", "bicubic", "trilinear", "conv"] = "conv",
            crop: bool = False
    ):
        super(UNet, self).__init__()

        # Make sure channels are given
        assert channels is not None, "channels must be specified"

        # Extract the number of layers from the list of channels
        num_layers = len(channels)

        # Prepare the layers vector
        self.downscale_layers = nn.ModuleList([])
        self.upscale_layers = nn.ModuleList([])

        # Create the first two convolutions
        self.input_conv = DoubleConvBlock(in_channels, channels[0], kernel_size=conv_kernel_size, padding=conv_padding)

        # Create the downscale layers
        for i in range(num_layers - 1):
            self.downscale_layers.append(
                DownscaleBlock(
                    in_channels=channels[i],
                    out_channels=channels[i + 1],
                    scale=scale,
                    pooling_type=pooling_type,
                    conv_kernel_size=conv_kernel_size,
                    conv_padding=conv_padding,
                )
            )

        # Create the upscale layers
        for i in range(num_layers - 1):
            self.upscale_layers.append(
                UpscaleConcatBlock(
                    in_channels=channels[-i - 1],
                    out_channels=channels[-i - 2],
                    scale=scale,
                    conv_kernel_size=conv_kernel_size,
                    conv_padding=conv_padding,
                    method=method,
                    crop=crop
                )
            )

        self.output_conv = nn.Conv2d(channels[0], out_channels, kernel_size=1)

    def forward(self, x):
        # Pass the input through the input convolutional layer
        x = self.input_conv(x)

        # Pass the input through the downscale layers
        identity = []
        for downscale_layer in self.downscale_layers:
            identity.append(x)
            x = downscale_layer(x)

        # Pass the input through the upscale layers
        for i, upscale_layer in enumerate(self.upscale_layers):
            x = upscale_layer(x, identity[-i - 1])

        # Pass the input through the output convolutional layer
        x = self.output_conv(x)

        return x


def main():
    unet = UNet(
        in_channels=1,
        out_channels=2,
        channels=[64, 128, 256, 512, 1024],
        pooling_type="max",
        scale=2,
        conv_kernel_size=3,
        conv_padding=0,
        method="conv",
        crop=True
    )
    x = torch.randn(1, 1, 572, 572)
    print(unet(x).shape)


if __name__ == "__main__":
    main()



from einops import rearrange
from torch import nn
from torch.nn import functional as F


class Discriminator(nn.Module):
    def __init__(self):
        """
        Initialize the Discriminator
        """

        super(Discriminator, self).__init__()
        self.conv1 = nn.Conv2d(1, 64, 3, padding=1)
        self.conv2 = nn.Conv2d(64, 128, 3, padding=1)
        self.conv3 = nn.Conv2d(128, 256, 3, padding=1)

        self.bn1 = nn.BatchNorm2d(64)
        self.bn2 = nn.BatchNorm2d(128)
        self.bn3 = nn.BatchNorm2d(256)

        self.fc1 = nn.Linear(2304, 1024)
        self.fc2 = nn.Linear(1024, 1)

        self.dropout = nn.Dropout(0.2)

    def forward(self, x):
        """
        Discriminator forward pass
        :param x: The input image
        :return: The probability that the input is real
        """

        x = F.relu(self.conv1(x))
        x = self.bn1(x)
        x = F.max_pool2d(x, 2, 2)

        x = F.relu(self.conv2(x))
        x = self.bn2(x)
        x = F.max_pool2d(x, 2, 2)

        x = F.relu(self.conv3(x))
        x = self.bn3(x)
        x = F.max_pool2d(x, 2, 2)

        # flatten the image using rearrange from einops
        x = rearrange(x, 'b ... -> b (...)')

        x = self.dropout(x)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = F.sigmoid(x)

        return x


from math import pi
from typing import List, Union

import torch
from einops import rearrange, repeat
from torch import nn


def _fourier_encode(x, max_freq, num_bands=4):
    x = x.unsqueeze(-1)
    device, dtype, orig_x = x.device, x.dtype, x

    scales = torch.linspace(1., max_freq / 2, num_bands, device=device, dtype=dtype)
    scales = scales[(*((None,) * (len(x.shape) - 1)), Ellipsis)]

    x = x * scales * pi
    x = torch.cat([x.sin(), x.cos()], dim=-1)
    x = torch.cat((x, orig_x), dim=-1)
    return x


class FourierEmbedding(nn.Module):
    """
    Fourier Embedding will append the fourier positional encoding to the given
    input.

    The fourier positional encoding is calculated on sines and cosines of the
    original input.

    @Params:
    num_freq_bands: int
        - determines how many sines and cosines to append

    max_freq: int
        - determines the frequency domain for sines and cosines

    constant_mapping: bool = False
        - determines if the mapping is constant or dynamic

    max_position: Union[List[int], int] = 1000
        - in case of constant mapping, this will be the constant for the linear space
        - it can be a list, having a maximum boundary for each dimension, excepting batch and last dim
        - it can be an int, and it will apply the same maximum for all dimensions

    For a given torch tensor with shape [... c]
    it will return a torch tensor of shape [... f]
    where f = c + input_axis * ((num_freq_bands * 2) + 1)
    """

    def __init__(
            self,
            max_freq: int,
            num_freq_bands: int,
            constant_mapping: bool = False,
            max_position: Union[List[int], int] = 1600
    ):
        """Initialize the class."""
        super(FourierEmbedding, self).__init__()
        self.max_freq = max_freq
        self.num_freq_bands = num_freq_bands
        self.constant_mapping = constant_mapping
        self.max_position = max_position

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Process the input to append num_freq_bands sines and cosines to it.
        """

        # Take important information about input
        b, *axis, _ = x.shape
        device = x.device
        dtype = x.dtype

        # Map the positions to a linear space in the interval [-1, 1], agnostic
        # of batch_size
        if self.constant_mapping:

            # Transform max_position into list
            if not isinstance(self.max_position, list):
                max_position = [self.max_position for _ in axis]
            else:
                assert len(self.max_position) == len(axis), \
                    f"Wrong number of dimensions were given to max_position. \n" \
                    f"Expected {len(axis)}, got {len(self.max_position)}."
                max_position = self.max_position

            # Make sure we have the right amount of space
            assert all(size <= max_pos for size, max_pos in zip(axis, max_position)), \
                f"More positions are necessary than what is configured.\n" \
                f"{axis} > {self.max_position}"

            axis_pos = [
                torch.linspace(-1.0, 1.0, steps=max_pos, device=device, dtype=dtype)[:size]
                for size, max_pos in zip(axis, max_position)
            ]
        else:
            axis_pos = list(map(
                lambda size: torch.linspace(-1.0, 1.0, steps=size, device=device, dtype=dtype),
                axis
            ))

        pos = torch.stack(torch.meshgrid(*axis_pos, indexing='ij'), dim=-1)

        # Apply the fourier encoding to calculate the sines and cosines
        enc_pos = _fourier_encode(pos, self.max_freq, self.num_freq_bands)

        # Reshape the content to be exactly same shape as input
        # (excepting last channel)
        enc_pos = rearrange(enc_pos, '... n d -> ... (n d)')

        # Repeat it for each element inside the batch
        enc_pos = repeat(enc_pos, '... -> b ...', b=b)

        # Concatenate the encoded positionals with the original channels
        return torch.cat([x, enc_pos], dim=-1)



from typing import Optional

import torch
from torch import nn
from torch.nn import functional as F

from vision_models_playground.components.convolutions.conv_block import ConvBlock


class BottleneckBlock(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            stride: int = 1,
            hidden_channels: Optional[int] = None,
    ):
        if hidden_channels is None:
            hidden_channels = out_channels

        super(BottleneckBlock, self).__init__()
        self.conv1 = ConvBlock(in_channels, hidden_channels, stride=stride, kernel_size=1, padding=0)
        self.conv2 = ConvBlock(hidden_channels, hidden_channels, stride=1)
        self.conv3 = ConvBlock(hidden_channels, out_channels, stride=1, apply_relu=False)

        self.identity_reshape = ConvBlock(
            in_channels, out_channels, stride=stride, apply_relu=False, kernel_size=1, padding=0
        ) if in_channels != out_channels else nn.Identity()

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)
        identity = self.identity_reshape(identity)
        out += identity
        out = F.relu(out)

        return out


def main():
    block = BottleneckBlock(in_channels=64, out_channels=128, hidden_channels=32, stride=1)

    x = torch.randn(1, 64, 32, 32)
    out = block(x)
    print(out.shape)


if __name__ == "__main__":
    main()



import torch
from torch import nn

from torchvision import datasets, transforms
from torchvision.utils import save_image

from vision_models_playground.models.generative.adverserial.mnist_discriminator import Discriminator
from vision_models_playground.models.generative.adverserial.mnist_generator import Generator


class GAN(nn.Module):
    def __init__(self, generator: nn.Module, discriminator: nn.Module):
        """
        Initialize the GAN.
        :param generator: the generator network
        :param discriminator: the discriminator network
        """

        super(GAN, self).__init__()
        self.generator = generator
        self.discriminator = discriminator

        beta1 = 0.5

        self.discriminator_optim = torch.optim.Adam(self.discriminator.parameters(), lr=0.0002, betas=(beta1, 0.999))
        self.generator_optim = torch.optim.Adam(self.generator.parameters(), lr=0.0002, betas=(beta1, 0.999))

        self.loss_func = nn.BCELoss()

        self.noise_dim = self.generator.noise_dim

    def forward(self, x: torch.Tensor):
        """
        Forward pass of the GAN.
        :param x: noise vector
        :return: generated images along with their predictions
        """

        img = self.generator(x)
        pred = self.discriminator(x)
        return img, pred

    def predict(self, x: torch.Tensor):
        """
        Predict the probability of the input being real.
        :param x: images to predict on
        :return: predicted probability
        """

        return self.discriminator.predict(x)

    def generate(self, noise: torch.Tensor):
        """
        Generate a set of samples from the generator, using the given noise
        :param noise: The noise to use to generate the samples
        :return: A list of generated images
        """

        return self.generator(noise)

    def generate_sample(self, num_instances: int = 10):
        """
        Generate a set of samples from the generator.
        :param num_instances: number of samples to generate
        :return: a list of generated samples
        """

        with torch.no_grad():
            z = torch.randn(num_instances, self.noise_dim).cuda()
            x_gen = self.generator(z)
            return x_gen

    def generate_quality_samples(self, num_instances: int = 10, step: int = 10):
        """
        Generate a set of quality samples from the generator, using the discriminator predictions.
        :param num_instances: number of samples to generate
        :param step: number of samples to generate at once, to avoid memory issues
        :return: a list of generated samples
        """

        # Init the results list with the proper shape
        generated_sample = self.generate_sample(1)
        batch_size, *generated_shape = generated_sample.shape
        results = torch.zeros(0, *generated_shape).cuda()

        with torch.no_grad():
            while results.shape[0] < num_instances:
                z = torch.randn(step, self.noise_dim).cuda()
                x_gen = self.generator(z)

                good = self.discriminator(x_gen).squeeze()

                mask = good > 0.5
                x_gen = x_gen[mask]
                results = torch.cat([results, x_gen])

            results = results[:num_instances]
            return results

    def train_step(self, x: torch.Tensor):
        """
        Train the GAN for one step.
        :param x: The real images to train on
        :return: The loss of the GAN
        """

        # zero the gradients
        self.discriminator_optim.zero_grad()
        self.generator_optim.zero_grad()

        # generate a batch of images
        z = torch.randn(x.shape[0], self.noise_dim).cuda()
        x_gen = self.generator(z)

        # train the discriminator
        d_loss = self.discriminator_loss(x, x_gen)
        d_loss.backward()
        self.discriminator_optim.step()

        # train the generator
        x_gen = self.generator(z)
        g_loss = self.generator_loss(x_gen)
        g_loss.backward()
        self.generator_optim.step()

        return d_loss, g_loss

    def discriminator_loss(self, x: torch.Tensor, x_gen: torch.Tensor):
        """
        Compute the discriminator loss.
        :param x: The real images
        :param x_gen: The generated images
        :return: The loss of the discriminator
        """

        # real loss
        y = torch.ones(x.shape[0]).cuda()
        y_pred = self.discriminator(x).squeeze()
        real_loss = self.loss_func(y_pred, y)

        # fake loss
        y = torch.zeros(x_gen.shape[0]).cuda()
        y_pred = self.discriminator(x_gen).squeeze()
        fake_loss = self.loss_func(y_pred, y)

        # total loss
        d_loss = real_loss + fake_loss
        return d_loss

    def generator_loss(self, x_gen: torch.Tensor):
        """
        Compute the generator loss.
        :param x_gen: The generated images
        :return: The loss of the generator
        """

        y = torch.ones(x_gen.shape[0]).cuda()
        y_pred = self.discriminator(x_gen).squeeze()
        g_loss = self.loss_func(y_pred, y)
        return g_loss

    def train_epochs(
            self,
            train_loader: torch.utils.data.DataLoader,
            epochs: int = 10,
            print_every: int = 100,
            save_checkpoint: bool = True,
            generate_images: bool = True
    ):
        """
        Train the GAN for a number of epochs.
        :param train_loader: The data loader for the training data
        :param epochs: The number of epochs to train for
        :param print_every: The number of steps to print the loss every
        :param save_checkpoint: Whether to save the model after each epoch
        :param generate_images: Whether to generate images after each print_every steps inside the epoch
        :return: None
        """

        for epoch in range(epochs):
            for i, x in enumerate(train_loader):
                if isinstance(x, tuple):
                    x = x[0]

                x = x.cuda()
                d_loss, g_loss = self.train_step(x)

                if i % print_every != 0:
                    continue

                print('Epoch: {}, Iteration: {}, D_Loss: {}, G_Loss: {}'.format(epoch, i, d_loss, g_loss))

                if not generate_images:
                    continue

                # generate a image
                with torch.no_grad():
                    z = torch.randn(1, self.noise_dim).cuda()
                    x_gen = self.generator(z)

                    # save the image
                    save_image(x_gen, 'images/{}.png'.format(i))

            # save the model
            if not save_checkpoint:
                continue

            torch.save(self.generator.state_dict(), 'models_checkpoints/generator.pt')
            torch.save(self.discriminator.state_dict(), 'models_checkpoints/discriminator.pt')


def main():
    # create GAN
    generator = Generator()
    discriminator = Discriminator()
    gan = GAN(generator, discriminator)

    # put model on cuda
    gan.cuda()

    # create the data loader
    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST('./data', train=True, download=True, transform=transforms.Compose([
            transforms.ToTensor(),
        ])),
        batch_size=64,
        shuffle=True
    )

    # train the GAN
    gan.train_epochs(train_loader, epochs=100, print_every=100)


if __name__ == '__main__':
    main()
